#include "../../sf.h"

#include "AudioManager.h"
#include "../../diag/assert.h"
#include "../../diag/log.h"
#include "../../misc/misc.h"
#include "../../misc/GamePack.h"
#include "../../misc/sys/time.h"
#include "../../misc/sys/filewrp.h"


namespace sf
{
namespace core
{

class JniHelpers
{
public:
    static void Init()
    {
        SFASSERT(!m_CreateMusic); // to prevent reinitialization
		SFASSERT(!m_CreateMusicfromGamePack); // to prevent reinitialization
		
        jclass media_player = JNIGetClass("android/media/MediaPlayer");
        m_JavaMediaManager = JNIGetClass("com/stargaze/sf/MediaManager");
        
        m_CreateMusic = JNIGetStaticMethod(m_JavaMediaManager, "createMusic",
                "(Ljava/lang/String;)Landroid/media/MediaPlayer;");
        m_CreateMusicfromGamePack = JNIGetStaticMethod(m_JavaMediaManager, "createMusicfromGamePack",
                "(II)Landroid/media/MediaPlayer;");
				
        m_MPStart = JNIGetObjectMethod(media_player, "start", "()V");
        m_MPPause = JNIGetObjectMethod(media_player, "pause", "()V");
        m_MPStop = JNIGetObjectMethod(media_player, "stop", "()V");
        m_MPIsPlaying = JNIGetObjectMethod(media_player, "isPlaying", "()Z");
        m_MPGetDuration = JNIGetObjectMethod(media_player, "getDuration", "()I");
        m_MPSetVolume = JNIGetObjectMethod(media_player, "setVolume", "(FF)V");
        m_MPSetLooping = JNIGetObjectMethod(media_player, "setLooping", "(Z)V");
        m_MPRelease = JNIGetObjectMethod(media_player, "release", "()V");
    }
        
    FORCEINLINE static jobject CreateMusic(const StringA120& _path)
    {
        jstring path = g_JavaEnv->NewStringUTF(_path.c_str());
        
        jobject music = g_JavaEnv->CallStaticObjectMethod(
            m_JavaMediaManager, m_CreateMusic, path);
            
        g_JavaEnv->DeleteLocalRef(path);
            
        if (music)
        {
            music = g_JavaEnv->NewGlobalRef(music);
            g_JavaEnv->DeleteLocalRef(music);
        }
            
        return music;
    }
	
	FORCEINLINE static jobject CreateMusicfromGamePack(size_t _offset, size_t _length)
    {
		jobject music = g_JavaEnv->CallStaticObjectMethod(
            m_JavaMediaManager, m_CreateMusicfromGamePack, _offset, _length);
        
        if (music)
        {
            music = g_JavaEnv->NewGlobalRef(music);
            g_JavaEnv->DeleteLocalRef(music);
        }
            
        return music;
    }
    
    FORCEINLINE static void PlayMusic(jobject _music)
    {
        g_JavaEnv->CallVoidMethod(_music, m_MPStart);
    }
    
    FORCEINLINE static void PauseMusic(jobject _music)
    {
        g_JavaEnv->CallVoidMethod(_music, m_MPPause);
    }
    
    FORCEINLINE static void StopMusic(jobject _music)
    {
        g_JavaEnv->CallVoidMethod(_music, m_MPStop);
    }
    
    FORCEINLINE static bool IsMusicPlaying(jobject _music)
    {
        return g_JavaEnv->CallBooleanMethod(_music, m_MPIsPlaying);
    }
    
    FORCEINLINE static int GetMusicDuration(jobject _music)
    {
        return g_JavaEnv->CallIntMethod(_music, m_MPGetDuration);
    }
    
    FORCEINLINE static void SetMusicVolume(jobject _music, float _volume)
    {
        g_JavaEnv->CallVoidMethod(_music, m_MPSetVolume, _volume, _volume);
    }
    
    FORCEINLINE static void SetMusicLooping(jobject _music, bool _looping)
    {
        g_JavaEnv->CallVoidMethod(_music, m_MPSetLooping, _looping);
    }
    
    FORCEINLINE static void ReleaseMusic(jobject _music)
    {
    	StopMusic(_music);
        g_JavaEnv->CallVoidMethod(_music, m_MPRelease);
        g_JavaEnv->DeleteGlobalRef(_music);
    }
    
    
private:
    static jclass m_JavaMediaManager;
    static jmethodID m_CreateMusic;
	static jmethodID m_CreateMusicfromGamePack;
    static jmethodID m_MPStart;
    static jmethodID m_MPPause;
    static jmethodID m_MPStop;
    static jmethodID m_MPIsPlaying;
    static jmethodID m_MPSetVolume;
    static jmethodID m_MPGetDuration;
    static jmethodID m_MPSetLooping;
    static jmethodID m_MPRelease;
};

jclass JniHelpers::m_JavaMediaManager = 0;
jmethodID JniHelpers::m_CreateMusic = 0;
jmethodID JniHelpers::m_CreateMusicfromGamePack= 0;
jmethodID JniHelpers::m_MPStart = 0;
jmethodID JniHelpers::m_MPPause = 0;
jmethodID JniHelpers::m_MPStop = 0;
jmethodID JniHelpers::m_MPIsPlaying = 0;
jmethodID JniHelpers::m_MPSetVolume = 0;
jmethodID JniHelpers::m_MPGetDuration = 0;
jmethodID JniHelpers::m_MPSetLooping = 0;
jmethodID JniHelpers::m_MPRelease = 0;

CAudioManager::CAudioManager() 
:	m_LastId(0),
	m_Volume(100),
	m_ParentTimer(sf::core::g_TimeManager::Instance().AttachGroupTimer("AudioManager"))
{
    JniHelpers::Init();
}

CAudioManager::~CAudioManager()
{	
	StopAll();
	UnloadAll();
}

void CAudioManager::SetGlobalVolume(float _volume)
{
	m_Volume = misc::clamp(_volume, MinVolume, MaxVolume);
}

float CAudioManager::GetGlobalVolume() const 
{
	return m_Volume;
}

void CAudioManager::AddGroup(const IdString &_group)
{
	m_GroupVolumeRepository.push_back( std::make_pair(_group, (float)MaxVolume));
}

bool CAudioManager::IsSound(const IdString &_id) const
{
	return FindSoundInfo(_id) != 0;
}

bool CAudioManager::IsSound(size_t _id) const
{
	int i = 0, e = m_Streams.Size();
	for(; i < e; ++i)
		if (m_Streams[i].playing_id == _id)
			return true;
	return false;
}

size_t CAudioManager::Play(const IdString &_id, float _balance, float _volume, int _priority, int _playing_count, float _speed)
{	
	if (_id.empty())
		return InvalidHandle;

	const SoundInfo *si = FindSoundInfo(_id);
	if (!si)
		return InvalidHandle;
	
	if (m_Streams.Size() == MaximumConcurrentStreams)
		return InvalidHandle; // we're full, just return
    
    PlayingSound ps;
    memset(&ps, 0, sizeof(ps));
    jobject music;
	
	// MediaPlayer
	if (si->offset == -1 || !(music = JniHelpers::CreateMusicfromGamePack(si->offset, si->length)))
		music = JniHelpers::CreateMusic(si->path);
	
	if (!music)
	{
		SFLOGA("sf", sf::diag::LogInfo, "Couldn't create MediaPlayer to play '%s'", _id.c_str());
		return InvalidHandle;
	}
	ps.track = music;
	ps.length = JniHelpers::GetMusicDuration(music);
    
        
	ps.info = si;
	ps.playing_volume = (_volume == DefaultValue ? si->volume : _volume);
	ps.playing_id = GetNextId();
	ps.repeat_count = (_playing_count == DefaultValue ? si->repeat : _playing_count);
	ps.played_count = 1;
	
	m_Streams.PushBack(ps);

	JniHelpers::SetMusicVolume(ps.track, CalcResultVolumeForStream(&ps) / 100.f);

	if (ps.repeat_count == UnlimitedRepeat)
		JniHelpers::SetMusicLooping(ps.track, true);
	

	JniHelpers::PlayMusic(ps.track);
    
	return ps.playing_id;
}

void CAudioManager::PauseAll(bool _pause)
{
	size_t i = 0, e = m_Streams.Size();
	for(;i<e;++i)
	{
		if (_pause)
			JniHelpers::PauseMusic(m_Streams[i].track);
		else
			JniHelpers::PlayMusic(m_Streams[i].track);
	}
}

bool CAudioManager::Stop(size_t _id, float _time)
{
	size_t i = 0, e = m_Streams.Size();
	for (; i < e; ++i)
		if (m_Streams[i].playing_id == _id)
		{
			StopPlayingSound(&m_Streams[i], _time);
			return true;
		}
}

void CAudioManager::StopPlayingSound(CAudioManager::PlayingSound *_sound, float _in_time)
{
	SFASSERT(_in_time >= 0.f);
	_sound->action_stop_start_time = m_ParentTimer->GetTime() + _in_time;
}
	
bool CAudioManager::Stop(const IdString &_id, float _time)
{
	size_t i = 0, e = m_Streams.Size();
	for (; i < e; ++i)
		if (m_Streams[i].info->id == _id)
		{
			StopPlayingSound(&m_Streams[i], _time);
			return true;
		}
	return false;
}

void CAudioManager::StopAll(float _time)
{
	size_t i = 0, e = m_Streams.Size();
	for(; i < e; ++i)
		StopPlayingSound(&m_Streams[i], _time);
}

void CAudioManager::StopGroup(const IdString &_group_id, float _time)
{
	size_t i = 0, e = m_Streams.Size();
	for (; i < e; ++i)
		if (m_Streams[i].info->group_template == _group_id)
			StopPlayingSound(&m_Streams[i], _time);
}

void CAudioManager::ChangeVolumeForStream(CAudioManager::PlayingSound *_sound, float _delta, float _time, float _shift_time) const
{
	SFASSERT( _time >= 1.f || _shift_time >= 1.f);

	_sound->action_volume_change_delta = _delta;
	_sound->action_volume_change_applied_delta = 0;
	_sound->action_volume_change_start_time = m_ParentTimer->GetTime() + _shift_time;
	_sound->action_volume_change_perform_time = _shift_time;
}

bool CAudioManager::SetVolume(size_t _id, float _volume, float _time, float _shift_time)
{
	size_t i = 0, e = m_Streams.Size();
	for(;i<e;++i)
		if (m_Streams[i].playing_id == _id)
		{
			if ( _time < 1.f && _shift_time < 1.f)
			{
				m_Streams[i].playing_volume = misc::clamp(_volume, core::CAudioManager::MinVolume, core::CAudioManager::MaxVolume);
                JniHelpers::SetMusicVolume(m_Streams[i].track, CalcResultVolumeForStream(&m_Streams[i]) / 100.f);
			}
			else
			{
				ChangeVolumeForStream(
									  &m_Streams[i],
									  misc::clamp(_volume, core::CAudioManager::MinVolume, core::CAudioManager::MaxVolume) - m_Streams[i].playing_volume,
									  _time,
									  _shift_time);				
			}
			return true;			
		}
	return false;
}

float CAudioManager::GetVolume(const IdString &_id) const
{
	const SoundInfo *si = FindSoundInfo(_id);
	return si ? si->volume : MinVolume;
}

float CAudioManager::GetVolume(size_t _id) const
{
	size_t i = 0, e = m_Streams.Size();
	for(;i<e;++i)
		if (m_Streams[i].playing_id == _id)
			return m_Streams[i].playing_volume;
			
	return MinVolume;
}

float CAudioManager::GetMaxVolume(size_t _id) const
{
	size_t i = 0, e = m_Streams.Size();
	for(;i<e;++i)
		if (m_Streams[i].playing_id == _id)
		{
			return m_Streams[i].info->volume;
		}
	return MinVolume;
}

void CAudioManager::SetGroupVolume(const IdString &_group, float _volume)
{	
	size_t i = 0, e = m_GroupVolumeRepository.size();
	for(;i<e;++i)
		if (m_GroupVolumeRepository[i].first == _group)
		{
			m_GroupVolumeRepository[i].second = _volume;
			return;
		}
	
	m_GroupVolumeRepository.push_back( std::make_pair(_group, _volume));
}

float CAudioManager::GetGroupVolume(const IdString &_group) const
{
	size_t i = 0, e = m_GroupVolumeRepository.size();
	for(;i<e;++i)
		if (m_GroupVolumeRepository[i].first == _group)
			return m_GroupVolumeRepository[i].second;
	return MaxVolume;
}

bool CAudioManager::SetPriority(const IdString &_id, int _priority){return true;}
bool CAudioManager::SetPriority(size_t _id, int _priority){return true;}
int CAudioManager::GetPriority(const IdString &_id) const{	return 1;}
int CAudioManager::GetPriority(size_t _id) const{return 1;}


double CAudioManager::GetLength(size_t _id) const
{
	size_t i = 0, e = m_Streams.Size();
	for(;i<e;++i)
		if (m_Streams[i].playing_id == _id)
		{
			return m_Streams[i].length;
		}	
	return 0;
}

bool CAudioManager::Update()
{
	int i = 0, e = m_Streams.Size();
	for(;i < e; ++i)
	{
		PlayingSound *ps = &m_Streams[i];

        if (!ps->is_dead && !ps->is_finished && ps->repeat_count >= 0 && !JniHelpers::IsMusicPlaying(ps->track))
        {
            ps->is_finished = true;
        }
		
		if (ps->action_stop_start_time) // stop sound on delayed action
			if (m_ParentTimer->GetTime() >= ps->action_stop_start_time)
			{
				JniHelpers::StopMusic(ps->track);
				ps->is_dead = true;
			}			
		
		// check if sound is already finished		
		if(!ps->is_dead && ps->is_finished)
		{
			if (ps->repeat_count == ps->played_count)
			{
				ps->is_dead = true;
			}
			else
			{
				ps->played_count++;
                
                JniHelpers::StopMusic(ps->track);
                JniHelpers::PlayMusic(ps->track);

                ps->is_finished = false;
			}

		}
		
		// TODO - add pause functionality
		
		if (ps->is_dead)
		{
			JniHelpers::ReleaseMusic(ps->track);
        
			// channel is dead - remove from active streams
			if (i < e - 1)
			{
				m_Streams[i] = m_Streams[e-1];
				m_Streams.Resize(e-1);
				--e;
				--i;
				continue;				
			}
			else
			{
				m_Streams.Resize(i);
				break;
			}			
		}
		

		// sliding volume change action
		if ( fabsf(ps->action_volume_change_delta) > 1.f )
		{
			if (m_ParentTimer->GetTime() >= ps->action_volume_change_start_time)
			{
				if (m_ParentTimer->GetTime() <= ps->action_volume_change_start_time + ps->action_volume_change_perform_time)
				{
					float dv = ps->action_volume_change_delta / ps->action_volume_change_perform_time;
					if (dv > 0)
					{
						if (dv + ps->action_volume_change_applied_delta > ps->action_volume_change_delta)
							dv = ps->action_volume_change_delta - ps->action_volume_change_applied_delta;
					}
					else
					{						
						if (dv + ps->action_volume_change_applied_delta < ps->action_volume_change_delta)
							dv = ps->action_volume_change_delta - ps->action_volume_change_applied_delta;
					}
					ps->playing_volume += dv;
					ps->action_volume_change_applied_delta += dv;
				}
				else
				{
					ps->playing_volume += ps->action_volume_change_delta - ps->action_volume_change_applied_delta;
					ps->action_volume_change_delta = 0;
				}

			}
		}
		
		
		// update Volume for this sound stream
		{
			float value = CalcResultVolumeForStream(ps) / 100.f;
			if (fabsf(ps->prev_bass_volume - value) > 0.01f)
			{
				JniHelpers::SetMusicVolume(ps->track, value);
				ps->prev_bass_volume = value;
			}
		}
	}

	return true;
}
	
size_t CAudioManager::GetPlayingNowSoundsAmount() const
{
	return m_Streams.Size();		
}

CGroupTimer* CAudioManager::GetParentTimer()
{
	return m_ParentTimer;
}

size_t CAudioManager::GetNextId()
{
	return ++m_LastId;
}
		
void CAudioManager::CreateSounds(const sound::CSoundDeclarationsLoader &_from)
{
	const sound::CSoundDeclarationsLoader::SingleStorageHead *ss = _from.GetData();

	for(size_t i = 0; i < ss->sounds_amount; ++i)
	{
		const sf::sound::CSoundDeclarationsLoader::SoundDesc &desc = ss->GetSounds()[i];

		SoundInfo si;
        
        bool add = false;
		si.offset = size_t(-1);
		if (g_GamePack::Instance().Access(desc.path.c_str()))
		{
			add = true;
			si.path = desc.path;
			si.offset = g_GamePack::Instance().GetFileOffset(desc.path.c_str(), &si.length);
		}
        
		if (add)
		{
			AddGroup(desc.group_template);
			
			si.id = desc.id;
			si.group_template = desc.group_template;
			si.volume = desc.volume;
			si.repeat = desc.repeat;

			m_SoundsInfo.push_back(si);
		}		
	}
	
	if (m_SoundsInfo.size()>1)
		std::sort(&m_SoundsInfo[0], &m_SoundsInfo[m_SoundsInfo.size()]);
}

const CAudioManager::SoundInfo *CAudioManager::FindSoundInfo(const IdString &_id) const
{
	// binary search
	int n1 = 0;
	int n2 = m_SoundsInfo.size() - 1;
	if (n2 > n1)
	{
		do
		{
			int m = (n1 + n2)/2;
			
			if (m_SoundsInfo[m].id == _id)
				return &m_SoundsInfo[m];
			if (m_SoundsInfo[m].id > _id)
				n2 = m - 1;
			else
				n1 = m + 1;
		} while(n1 <= n2);
	}
	
	if (n1 >= 0 && n1 < (int)m_SoundsInfo.size() && m_SoundsInfo[n1].id == _id)
		return &m_SoundsInfo[n1];	
	
	return 0;		
}
	
float CAudioManager::CalcResultVolumeForStream(const CAudioManager::PlayingSound *_sound) const
{
	//   
	float result_volume = _sound->playing_volume;
	
	//	       
	result_volume *= GetGroupVolume( _sound->info->group_template ) / (float)(MaxVolume - MinVolume);

	//	        
	result_volume *= GetGlobalVolume() / (float)(MaxVolume - MinVolume);
	 
	SFASSERT(result_volume >= MinVolume);
	SFASSERT(result_volume <= MaxVolume);
	return result_volume;
}
	
} //sound

} //sf
